@@ -1,4 +1,4 @@
-/* $Id: milter-spamd.c,v 1.6 2004/04/04 18:23:55 dhartmei Exp $ */
+/* $Id: milter-spamd.c,v 1.1.1.1 2007/01/11 15:53:57 dhartmei Exp $ */
 
 /*
  * Copyright (c) 2004 Daniel Hartmeier
@@ -30,7 +30,7 @@
  *
  */
 
-static const char rcsid[] = "$Id: milter-spamd.c,v 1.6 2004/04/04 18:23:55 dhartmei Exp $";
+static const char rcsid[] = "$Id: milter-spamd.c,v 1.1.1.1 2007/01/11 15:53:57 dhartmei Exp $";
 
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -55,16 +55,19 @@ static const char	*spamd_user = "";
 static const char	*ignore_connect = "";
 static int		 debug = 0;
 static regex_t		 re_ignore_connect;
-static pthread_mutex_t	 mutex;
 
 struct context {
 	char		 buf[2048];
 	unsigned	 pos;
+	char		 host[128];
 	char		 addr[64];
+	char		 helo[128];
+	char		 env_rcpt[128];
 	char		 hdr_from[128];
 	char		 hdr_to[128];
 	char		 hdr_subject[128];
 	int		 fd;
+	unsigned	 lines;
 	int		 state;
 	int		 spam;
 	double		 score, threshold;
@@ -83,14 +86,13 @@ static sfsistat		 cb_abort(SMFICTX *);
 static sfsistat		 cb_close(SMFICTX *);
 static void		 usage(const char *);
 static void		 msg(int, struct context *, const char *, ...);
-static void		 mutex_lock(void);
-static void		 mutex_unlock(void);
-static void		 die(const char *);
 
 #define USER		"_milter-spamd"
 #define OCONN		"unix:/var/spool/milter-spamd/sock"
 #define RCODE_REJECT	"554"
 #define XCODE_REJECT	"5.7.1"
+#define	SPAMD_ADDR	"127.0.0.1"
+#define	SPAMD_PORT	783
 
 #define	ST_MTIME st_mtimespec
 
@@ -107,8 +109,8 @@ get_spamd_fd(struct context *context)
 	}
 	memset(&sa, 0, sizeof(sa));
 	sa.sin_family = AF_INET;
-	sa.sin_addr.s_addr = inet_addr("127.0.0.1");
-	sa.sin_port = htons(783);
+	sa.sin_addr.s_addr = inet_addr(SPAMD_ADDR);
+	sa.sin_port = htons(SPAMD_PORT);
 	if (connect(fd, (struct sockaddr *)&sa, sizeof(sa))) {
 		msg(LOG_ERR, context, "get_spamd_fd: connect: %s",
 		    strerror(errno));
@@ -137,26 +139,29 @@ cb_connect(SMFICTX *ctx, char *name, _SOCK_ADDR *sa)
 	}
 
 	strlcpy(host, "unknown", sizeof(host));
-	switch (sa->sa_family) {
-	case AF_INET: {
-		struct sockaddr_in *sin = (struct sockaddr_in *)sa;
+	if (sa) {
+		switch (sa->sa_family) {
+		case AF_INET: {
+			struct sockaddr_in *sin = (struct sockaddr_in *)sa;
 
-		if (inet_ntop(AF_INET, &sin->sin_addr.s_addr, host,
-		    sizeof(host)) == NULL)
-			msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
-			    strerror(errno));
-		break;
-	}
-	case AF_INET6: {
-		struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
+			if (inet_ntop(AF_INET, &sin->sin_addr.s_addr, host,
+			    sizeof(host)) == NULL)
+				msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
+				    strerror(errno));
+			break;
+		}
+		case AF_INET6: {
+			struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
 
-		if (inet_ntop(AF_INET6, &sin6->sin6_addr, host,
-		    sizeof(host)) == NULL)
-			msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
-			    strerror(errno));
-		break;
-	}
+			if (inet_ntop(AF_INET6, &sin6->sin6_addr, host,
+			    sizeof(host)) == NULL)
+				msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
+				    strerror(errno));
+			break;
+		}
+		}
 	}
+	strlcpy(context->host, name, sizeof(context->host));
 	strlcpy(context->addr, host, sizeof(context->addr));
 	msg(LOG_DEBUG, context, "cb_connect('%s', '%s')", name, host);
 	if (ignore_connect[0] && (
@@ -177,6 +182,7 @@ cb_helo(SMFICTX *ctx, char *arg)
 		msg(LOG_ERR, NULL, "cb_helo: smfi_getpriv");
 		return (SMFIS_ACCEPT);
 	}
+	strlcpy(context->helo, arg, sizeof(context->helo));
 	msg(LOG_DEBUG, context, "cb_helo('%s')", arg);
 	return (SMFIS_CONTINUE);
 }
@@ -204,8 +210,10 @@ cb_envrcpt(SMFICTX *ctx, char **args)
 		msg(LOG_ERR, NULL, "cb_envrcpt: smfi_getpriv");
 		return (SMFIS_ACCEPT);
 	}
-	if (*args != NULL)
+	if (*args != NULL) {
+		strlcpy(context->env_rcpt, *args, sizeof(context->env_rcpt));
 		msg(LOG_DEBUG, context, "cb_envrcpt('%s')", *args);
+	}
 	return (SMFIS_CONTINUE);
 }
 
@@ -231,19 +239,41 @@ cb_header(SMFICTX *ctx, char *name, char *value)
 		return (SMFIS_ACCEPT);
 	}
 	msg(LOG_DEBUG, context, "cb_header('%s', '%s')", name, value);
-	mutex_lock();
 	if (context->fd < 0) {
-		if ((context->fd = get_spamd_fd(context)) < 0) {
-			mutex_unlock();
+		const char *sendmail_name = smfi_getsymval(ctx, "j");
+		const char *sendmail_queue = smfi_getsymval(ctx, "i");
+		const char *sendmail_date = smfi_getsymval(ctx, "b");
+
+		if ((context->fd = get_spamd_fd(context)) < 0)
 			return (SMFIS_ACCEPT);
-		}
 		fdprintf(context->fd, "SYMBOLS SPAMC/1.2\r\n");
 		if (spamd_user[0])
 			fdprintf(context->fd, "User: %s\r\n", spamd_user);
 		fdprintf(context->fd, "\r\n");
+		/* send fake Received: header */
+		fdprintf(context->fd, "Received: from %s (%s [%s])",
+		    context->helo, context->host, context->addr);
+		if (sendmail_name != NULL && sendmail_name[0]) {
+			fdprintf(context->fd, "\r\n\tby %s (milter-spamd)",
+			    sendmail_name);
+			if (sendmail_queue != NULL && sendmail_queue[0])
+				fdprintf(context->fd, " id %s", sendmail_queue);
+		}
+		if (context->env_rcpt[0])
+			fdprintf(context->fd, "\r\n\tfor %s", context->env_rcpt);
+		if (sendmail_date != NULL && sendmail_date[0])
+			fdprintf(context->fd, "; %s", sendmail_date);
+		else {
+			char d[128];
+			time_t t = time(NULL);
+
+			if (strftime(d, sizeof(d), "%a, %e %b %Y %H:%M:%S %z",
+			    localtime(&t)))
+				fdprintf(context->fd, "; %s", d);
+		}
+		fdprintf(context->fd, "\r\n");
 	}
 	fdprintf(context->fd, "%s: %s\r\n", name, value);
-	mutex_unlock();
 	if (!strcasecmp(name, "From"))
 		strlcpy(context->hdr_from, value, sizeof(context->hdr_from));
 	else if (!strcasecmp(name, "To"))
@@ -264,10 +294,8 @@ cb_eoh(SMFICTX *ctx)
 		return (SMFIS_ACCEPT);
 	}
 	msg(LOG_DEBUG, context, "cb_eoh()");
-	mutex_lock();
 	if (context->fd >= 0)
 		fdprintf(context->fd, "\r\n");
-	mutex_unlock();
 	memset(context->buf, 0, sizeof(context->buf));
 	context->pos = 0;
 	return (SMFIS_CONTINUE);
@@ -282,7 +310,6 @@ cb_body(SMFICTX *ctx, u_char *chunk, size_t size)
 		msg(LOG_ERR, NULL, "cb_body: smfi_getpriv");
 		return (SMFIS_ACCEPT);
 	}
-	mutex_lock();
 	for (; size > 0; size--, chunk++) {
 		context->buf[context->pos] = *chunk;
 		if (context->buf[context->pos] == '\n' ||
@@ -294,12 +321,13 @@ cb_body(SMFICTX *ctx, u_char *chunk, size_t size)
 				context->buf[context->pos] = 0;
 			context->pos = 0;
 			msg(LOG_DEBUG, context, "cb_body('%s')", context->buf);
-			if (context->fd >= 0)
+			if (context->fd >= 0 && context->lines < 1000) {
 				fdprintf(context->fd, "%s\r\n", context->buf);
+				context->lines++;
+			}
 		} else
 			context->pos++;
 	}
-	mutex_unlock();
 	return (SMFIS_CONTINUE);
 }
 
@@ -373,9 +401,9 @@ cb_eom(SMFICTX *ctx)
 		return (SMFIS_ACCEPT);
 	}
 	msg(LOG_DEBUG, context, "cb_eom()");
-	mutex_lock();
 	if (context->fd < 0)
 		goto done;
+	context->symbols[0] = 0;
 	/* no more writing data to spamd, want to read result now */
 	if (shutdown(context->fd, SHUT_WR)) {
 		msg(LOG_ERR, context, "cb_eom: shutdown: %s", strerror(errno));
@@ -386,16 +414,16 @@ cb_eom(SMFICTX *ctx)
 		msg(LOG_ERR, context, "cb_eom: fcntl: %s", strerror(errno));
 		goto done;
 	}
-	/* try at most 10 times (2 seconds timeout each) */
-	while (action == SMFIS_CONTINUE && retry < 10) {
+	/* try at most 6 times (10 seconds timeout each) */
+	while (action == SMFIS_CONTINUE && retry < 6) {
 		fd_set fds;
 		struct timeval tv;
 		int r, i;
-		char b[2048];
+		char b[8192];
 
 		FD_ZERO(&fds);
 		FD_SET(context->fd, &fds);
-		tv.tv_sec = 2;
+		tv.tv_sec = 10;
 		tv.tv_usec = 0;
 		r = select(context->fd + 1, &fds, NULL, NULL, &tv);
 		if (r < 0) {
@@ -407,7 +435,7 @@ cb_eom(SMFICTX *ctx)
 			continue;
 		} else if (r == 0 || !FD_ISSET(context->fd, &fds)) {
 			retry++;
-			msg(LOG_NOTICE, context, "cb_eom: waiting for "
+			msg(LOG_DEBUG, context, "cb_eom: waiting for "
 			    "spamd reply (retry %d)", retry);
 			continue;
 		}
@@ -434,7 +462,7 @@ cb_eom(SMFICTX *ctx)
 			} else
 				buf[pos++] = b[i];
 	}
-	if (retry == 10)
+	if (retry == 6)
 		msg(LOG_ERR, context, "cb_eom: spamd connection timed out");
 done:
 	if (context->fd >= 0) {
@@ -459,9 +487,9 @@ done:
 			msg(LOG_ERR, context, "smfi_setreply");
 	}
 	context->pos = context->hdr_from[0] = context->hdr_to[0] =
-	    context->hdr_subject[0] = context->state = context->spam = 0;
+	    context->hdr_subject[0] = context->state = context->spam =
+	    context->symbols[0] = 0;
 	context->score = context->threshold = 0.0;
-	mutex_unlock();
 	return (action);
 }
 
@@ -475,8 +503,13 @@ cb_abort(SMFICTX *ctx)
 		return (SMFIS_ACCEPT);
 	}
 	msg(LOG_DEBUG, context, "cb_abort()");
+	if (context->fd >= 0) {
+		close(context->fd);
+		context->fd = -1;
+	}
 	context->pos = context->hdr_from[0] = context->hdr_to[0] =
-	    context->hdr_subject[0] = context->state = context->spam = 0;
+	    context->hdr_subject[0] = context->state = context->spam =
+	    context->symbols[0] = 0;
 	context->score = context->threshold = 0.0;
 	return (SMFIS_CONTINUE);
 }
@@ -488,7 +521,6 @@ cb_close(SMFICTX *ctx)
 
 	context = (struct context *)smfi_getpriv(ctx);
 	msg(LOG_DEBUG, context, "cb_close()");
-	mutex_lock();
 	if (context != NULL) {
 		smfi_setpriv(ctx, NULL);
 		if (context->fd >= 0) {
@@ -497,7 +529,6 @@ cb_close(SMFICTX *ctx)
 		}
 		free(context);
 	}
-	mutex_unlock();
 	return (SMFIS_CONTINUE);
 }
 
@@ -615,10 +646,6 @@ main(int argc, char **argv)
 		fprintf(stderr, "smfi_register: failed\n");
 		goto done;
 	}
-	if (pthread_mutex_init(&mutex, 0)) {
-		fprintf(stderr, "pthread_mutex_init: failed\n");
-		goto done;
-	}
 
 	/* daemonize (detach from controlling terminal) */
 	if (!debug && daemon(0, 0)) {
@@ -639,30 +666,6 @@ done:
 	return (r);
 }
 
-static void
-mutex_lock(void)
-{
-	if (pthread_mutex_lock(&mutex))
-		die("pthread_mutex_lock");
-}
-
-static void
-mutex_unlock(void)
-{
-	if (pthread_mutex_unlock(&mutex))
-		die("pthread_mutex_unlock");
-}
-
-static void
-die(const char *reason)
-{
-	msg(LOG_ERR, NULL, "die: %s", reason);
-	smfi_stop();
-	sleep(60);
-	/* not reached, smfi_stop() kills thread */
-	abort();
-}
-
 static void
 msg(int priority, struct context *context, const char *fmt, ...)
 {
